from typing import cast, Dict, Optional, Tuple, TYPE_CHECKING
import numpy as np
from scipy import signal

from .. import Model
from ..utils import Array, ArrayLike


if TYPE_CHECKING:
    num_samples = float
    num_y_channels = float
    num_values = float
    num_spectrum_pts = float
    num_spectrum_samples = float


def param_min_sqrs(
    model: Model,
    x: Array[("num_samples",), np.float64],
    y: Array[("num_y_channels", "num_samples"), np.float64],
    scanned_param: str,
    scanned_param_values: ArrayLike["num_values", np.float64],
    defaults: Optional[Dict[str, float]] = None,
) -> Tuple[float, float]:
    """Scans one model parameter while holding the others fixed to find the value
    that gives the best fit to the data (minimum sum-squared residuals).

    :param x: x-axis data
    :param y: y-axis data
    :param scanned_param: name of parameter to optimize
    :param scanned_param_values: array of scanned parameter values to test
    :param defaults: optional dictionary of fallback values to use for non-scanned
      parameters which don't have heuristics set yet

    :returns: tuple with the value from :param scanned_param_values: which results
    in lowest residuals and the root-sum-squared residuals for that value.
    """
    defaults = cast(dict, defaults or {})
    if not set(defaults.keys()).issubset(model.parameters.keys()):
        raise ValueError("Defaults must be a subset of the model parameters")
    if scanned_param in defaults.keys():
        raise ValueError("The scanned parameter cannot have a default")

    missing_values = [
        param_name
        for param_name, param_data in model.parameters.items()
        if not param_data.has_initial_value()
        and param_name != scanned_param
        and param_name not in defaults.keys()
    ]
    if any(missing_values):
        raise ValueError(f"No initial value specified for parameters: {missing_values}")

    param_values = {
        param: param_data.get_initial_value(default=defaults.get(param))
        for param, param_data in model.parameters.items()
        if param != scanned_param
    }

    scanned_param_values = np.asarray(scanned_param_values).squeeze()
    costs = np.zeros_like(scanned_param_values)
    for idx, value in np.ndenumerate(scanned_param_values):
        param_values[scanned_param] = value
        y_params = model.func(x, param_values)
        costs[idx] = np.sqrt(np.sum(np.square(y - y_params)))

    # handle a quirk of numpy indexing if only one value is passed in
    if scanned_param_values.size == 1:
        return float(scanned_param_values), float(costs)

    opt = np.argmin(costs)
    return float(scanned_param_values[opt]), float(costs[opt])


def get_sym_x(
    x: Array[("num_samples",), np.float64],
    y: Array[("num_y_channels", "num_samples"), np.float64],
) -> float:
    """Returns `x_0` such that y(x-x_0) is maximally symmetric."""
    y = np.atleast_2d(y)
    x_span = x.ptp()
    num_samples = x.size
    window_min = min(x) + 0.125 * x_span
    window_max = max(x) - 0.125 * x_span

    # min of three points per window
    window_min = max(window_min, x[2])
    window_max = min(window_max, x[-3])

    window = np.argwhere(np.logical_and(x >= window_min, x <= window_max)).squeeze()
    costs = np.zeros_like(window, dtype=float)

    for point_num, idx in np.ndenumerate(window.squeeze()):
        samples_left = idx
        samples_right = num_samples - idx - 1
        num_samples_idx = min(samples_left, samples_right)
        idx_left = idx - num_samples_idx
        idx_right = idx + num_samples_idx

        samples_left = y[:, idx_left:idx]
        samples_right = y[:, (idx + 1) : (idx_right + 1)]

        assert samples_left.shape == samples_right.shape

        diff = samples_left - np.fliplr(samples_right)

        # give more weight to windows with more structure
        mu = np.mean(samples_left)
        var = np.sum(np.square(samples_left - mu)) / samples_left.size

        cost = np.sqrt(np.sum(np.square(diff))) / samples_left.size
        costs[point_num] = cost / var

    return x[window[np.argmin(costs)]]


def find_x_offset_fft(
    x: Array[("num_samples",), np.float64],
    omega: Array[("num_spectrum_pts",), np.float64],
    spectrum: Array[("num_spectrum_pts",), np.float64],
    omega_cut_off: float,
) -> float:
    """Finds the x-axis offset of a dataset from the phase of an FFT.

    This function uses the FFT shift theorem to extract the offset from the phase
    slope of an FFT. At present it only supports models with a single y channel.

    :param omega: FFT frequency axis
    :param spectrum: complex FFT data. For models with multiple y channels, this
      should contain data from a single channel only.
    :param omega_cut_off: highest value of omega to use in offset estimation

    :returns: an estimate of the x-axis offset
    """
    if spectrum.ndim != 1:
        raise ValueError(f"Function only takes 1 y channel, not {spectrum.shape[1]}")

    keep = omega < omega_cut_off
    if np.sum(keep) < 2:
        raise ValueError("Insufficient data below cut-off")

    omega = omega[keep]
    phi = np.unwrap(np.angle(spectrum[keep]))
    phi -= phi[0]

    p = np.polyfit(omega, phi, deg=1)

    x0 = min(x) - p[0]
    x0 = x0 if x0 > min(x) else x0 + x.ptp()
    return x0


def find_x_offset_sym_peak_fft(
    model: Model,
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples",), np.float64],
    omega: Array[("num_spectrum_pts",), np.float64],
    spectrum: Array[("num_spectrum_pts",), np.float64],
    omega_cut_off: float,
    test_pts: Optional[Array[("num_values",), np.float64]] = None,
    x_offset_param_name: str = "x0",
    y_offset_param_name: str = "y0",
    defaults: Optional[Dict[str, float]] = None,
):
    """Finds the x-axis offset for symmetric, peaked (maximum deviation from the
    baseline occurs at the origin) functions.

    This heuristic draws candidate x-offset points from three sources and picks the
    best one (in the least-squares residuals sense). Sources:
      - FFT shift theorem based on provided spectrum data
      - Tests all points in the top quartile of deviation from the baseline
      - Optionally, user-provided "test points", taken from another heuristic. This
        allows the developer to combine the general-purpose heuristics here with
        other heuristics which make use of more model-specific assumptions

    :param model: the fit model
    :param x: x-axis data
    :param y: y-axis data. For models with multiple y channels, this should contain
        data from a single channel only.
    :param omega: FFT frequency axis
    :param spectrum: complex FFT data. For models with multiple y channels, this
      should contain data from a single channel only.
    :param omega_cut_off: highest value of omega to use in offset estimation
    :param test_pts: optional array of x-axis points to test
    :param x_offset_param_name: name of the x-axis offset model parameter
    :param y_offset_param_name: name of the y-axis offset model parameter
    :param defaults: optional dictionary of fallback values to use for parameters with
        no initial value specified.

    :returns: an estimate of the x-axis offset
    """
    defaults = defaults or {}

    if y.ndim != 1:
        raise ValueError(
            f"{y.shape[0]} y-channels were provided to a method which takes 1."
        )

    x0_candidates = np.array([])

    if test_pts is not None:
        x0_candidates = np.append(x0_candidates, test_pts)

    try:
        fft_candidate = find_x_offset_fft(
            x=x, omega=omega, spectrum=spectrum, omega_cut_off=omega_cut_off
        )
        x0_candidates = np.append(x0_candidates, fft_candidate)
    except ValueError:
        pass

    y0_default = defaults.get(y_offset_param_name)
    y0 = model.parameters[y_offset_param_name].get_initial_value(y0_default)
    deviations = np.argsort(np.abs(y - y0))
    top_quartile_deviations = deviations[int(len(deviations) * 3 / 4) :]
    deviations_candidates = x[top_quartile_deviations]
    x0_candidates = np.append(x0_candidates, deviations_candidates)

    best_x0, _ = param_min_sqrs(
        model=model,
        x=x,
        y=y,
        scanned_param=x_offset_param_name,
        scanned_param_values=x0_candidates,
        defaults=defaults,
    )
    return best_x0


def get_spectrum(
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples",), np.float64],
    density_units: bool = True,
    trim_dc: bool = False,
) -> Tuple[
    Array[("num_spectrum_samples",), np.float64],
    Array[("num_spectrum_samples",), np.float64],
]:
    """Returns the frequency spectrum (Fourier transform) of a dataset.

    :param x: 1D ndarray of shape (num_samples,) containing x-axis data
    :param y: 1D ndarray of shape (num_samples,) containing y-axis data
    :param density_units: if `False` we apply normalization for narrow-band signals. If
        `True` we normalize for continuous distributions.
    :param trim_dc: if `True` we do not return the DC component.
    """
    if x.ndim != 1:
        raise ValueError("x-axis data must be a 1D array.")
    if y.ndim != 1:
        raise ValueError("y-axis data must be a 1D array.")

    dx = x.ptp() / x.size
    n = x.size
    omega = np.fft.fftfreq(n, dx) * (2 * np.pi)
    y_f = np.fft.fft(y) * (x.ptp() / n)

    y_f = y_f[: int(n / 2)]
    omega = omega[: int(n / 2)]

    if density_units:
        d_omega = 2 * np.pi / (n * dx)
        y_f /= d_omega

    if trim_dc:
        omega = omega[1:]
        y_f = y_f[1:]

    return omega, y_f


def get_pgram(
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples",), np.float64],
    density_units: bool = True,
) -> Tuple[
    Array[("num_spectrum_samples",), np.float64],
    Array[("num_spectrum_samples",), np.float64],
]:
    """Returns a periodogram for a dataset, converted into amplitude units.

    Based on the Lombe-Scargle periodogram (essentially least-squares fitting of
    sinusoids at different frequencies).

    :param x: x-axis data
    :param y: y-axis data. For models with multiple y channels, this should contain
        data from a single channel only.
    :param density_units: if `False` (default) we apply normalization for narrow-band
        signals. If `True` we normalize for continuous distributions.
    :returns: tuple with the frequency axis (angular units) and the periodogram
    """
    if y.ndim != 1 and y.shape[1] > 1:
        raise ValueError(
            f"{y.shape[1]} y channels were provided to a method which takes 1"
        )

    dx = np.min(np.diff(x))
    duration = x.ptp()
    n = int(duration / dx)
    d_omega = 2 * np.pi / (n * dx)

    # Nyquist limit does not apply to irregularly spaced data
    # We'll use it as a starting point anyway...
    f_nyquist = 0.5 / dx

    omega_list = 2 * np.pi * np.linspace(d_omega, f_nyquist, n)
    pgram = signal.lombscargle(x, y, omega_list, precenter=True)
    pgram = np.sqrt(np.abs(pgram) * 4 / len(y))

    if density_units:
        pgram /= d_omega

    return omega_list, pgram


def find_x_offset_sampling(
    model: Model,
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples", "num_y_channels"), np.float64],
    width: float,
    x_offset_param_name: str = "x0",
) -> float:
    """Finds the x-axis offset of a dataset by stepping through a range of potential
    offset values and picking the one that gives the lowest residuals.

    This function takes a more brute-force approach by evaluating the model at a
    range of offset values, picking the one that gives the lowest residuals. This
    may be appropriate where one needs the estimate to be highly robust in the face
    of noisy, irregularly sampled data.

    :param x: x-axis data
    :param y: y-axis data
    :param width: width of the feature we're trying to find (e.g. FWHMH). Used to
        pick the spacing between offset values to try.
    :param x_offset_param_name: name of the x-axis offset parameter

    :returns: an estimate of the x-axis offset
    """
    offsets = np.arange(min(x), max(x), width / 6)
    return param_min_sqrs(
        model=model,
        x=x,
        y=y,
        scanned_param=x_offset_param_name,
        scanned_param_values=offsets,
    )[0]



from typing import Tuple, TYPE_CHECKING
import numpy as np

from . import heuristics, Triangle
from .. import common, Model, ModelParameter, NormalFitter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float
    num_y_channels = float


class ConeSlice(Model):
    """Slice through a cone.

    We parametrise cones as:
        z = sign * sqrt( (k_x * (x - x0))**2 + (k_y * (y - y0)) ** 2)) + z0

    This model represents a slice through the cone with fixed `y`, given by:
        z = sign(k_x) * sqrt( (k_x * (x - x0))**2 + alpha ** 2 ) + z0
    where:
      - alpha = k_y * (y - y0)
      - we use the sign of k_x to set the sign for the cone

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - z0: vertical offset to the cone. Fixed to 0 by default
      - alpha: offset due to being off-centre in the y-axis
      - k: slope along x

    Floating z0 and alpha without a user-estimate for either may result in an unreliable
    fit.
    """

    def get_num_y_channels(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[bool, bool]:
        return False, False

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x0: ModelParameter(scale_func=common.scale_no_rescale),
        z0: ModelParameter(scale_func=common.scale_no_rescale, fixed_to=0),
        k: ModelParameter(scale_func=common.scale_no_rescale),
        alpha: ModelParameter(lower_bound=0, scale_func=common.scale_no_rescale),
    ) -> Array[("num_samples",), np.float64]:
        return np.sign(k) * np.sqrt((k * (x - x0)) ** 2 + alpha**2) + z0

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
    ):
        y = y.squeeze()

        fit = NormalFitter(x=x, y=y, model=Triangle())
        self.parameters["x0"].heuristic = fit.values["x0"]
        self.parameters["k"].heuristic = k = fit.values["k"]

        peak_value = fit.values["y0"]  # peak_value = alpha * k + z0

        if self.parameters["z0"].has_user_initial_value():
            z0 = self.parameters["z0"].get_initial_value()
            self.parameters["alpha"].heuristic = (peak_value - z0) / k
        elif self.parameters["alpha"].has_user_initial_value():
            alpha = self.parameters["alpha"].get_initial_value()
            self.parameters["z0"].heuristic = peak_value - alpha * k
        else:  # Hope one of z0 / alpha dominates the other
            _, z0_cost = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="z0",
                scanned_param_values=[peak_value],
                defaults={"alpha": 0},
            )
            _, alpha_cost = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="alpha",
                scanned_param_values=[peak_value / k],
                defaults={"z0": 0},
            )
            if z0_cost < alpha_cost:
                self.parameters["z0"].heuristic = peak_value
                self.parameters["alpha"].heuristic = 0
            else:
                self.parameters["z0"].heuristic = 0
                self.parameters["alpha"].heuristic = peak_value



from __future__ import annotations

import copy
import dataclasses
import inspect
import logging
import numpy as np
from typing import Callable, Dict, List, Optional, Tuple, Union, TYPE_CHECKING

from .utils import Array, ArrayLike


if TYPE_CHECKING:
    num_samples = float
    num_values = float
    num_spectrum_pts = float
    num_test_pts = float
    num_y_channels = float


logger = logging.getLogger(__name__)


TX = ArrayLike[("num_samples",), np.float64]
TY = Union[
    Array[("num_y_channels", "num_samples"), np.float64],
    Array[("num_samples"), np.float64],
]


def scale_invariant(x_scale: float, y_scale: float) -> float:
    """Scale function for model parameters whose value is invariant under rescaling of
    the x- and y-axes"""
    return 1


def scale_x(x_scale: float, y_scale: float) -> float:
    """
    Scale function for model parameters whose value scales linearly with the x-axis
    """
    return x_scale


def scale_x_inv(x_scale: float, y_scale: float) -> float:
    """
    Scale function for model parameters whose value scales with 1/x_scale
    """
    return 1 / x_scale


def scale_y(x_scale: float, y_scale: float) -> float:
    """
    Scale function for model parameters whose value scales linearly with the x-axis
    """
    return y_scale


def scale_power(x_power: int, y_power: int) -> Callable[[float, float], float]:
    """Returns a scale function for model parameters whose value scales as a power of
    the x- and y-axes
    """

    def scale_func(x_scale: float, y_scale: float) -> float:
        return (x_scale**x_power) * (y_scale**y_power)

    return scale_func


def scale_undefined(x_scale: float, y_scale: float) -> float:
    """This is typically used when the appropriate scale factor to use must be
    determined at runtime"""
    raise RuntimeError(
        "Attempt to rescale model parameter with undefined scale function"
    )


def scale_no_rescale(x_scale: float, y_scale: float) -> float:
    """For model parameters which cannot be rescaled"""
    if x_scale != 1.0 or y_scale != 1:
        raise RuntimeError(
            "Attempt to rescale model parameter which does not support rescaling"
        )
    return 1.0


@dataclasses.dataclass
class ModelParameter:
    """Metadata associated with a model parameter.

    Attributes:
        scale_func: callable returning a scale factor which the parameter must be
            *multiplied* by if it was fitted using `x` / `y` data that has been
            *multiplied* by the given scale factors. Scale factors are used to improve
            numerical stability by avoiding asking the optimizer to work with very large
            or very small values of `x` and `y`. The callable takes the x-axis and
            y-axis scale factors as arguments. A number of default scale functions are
            provided for convenience.
        lower_bound: lower bound for the parameter. Fitted values are guaranteed to be
            greater than or equal to the lower bound. Parameter bounds may be used by
            fit heuristics to help find good starting points for the optimizer.
        upper_bound: upper bound for the parameter. Fitted values are guaranteed to be
            lower than or equal to the upper bound. Parameter bounds may be used by
            fit heuristics to help find good starting points for the optimizer.
        fixed_to: if not `None`, the model parameter is fixed to this value during
            fitting instead of being floated. This value may additionally be used by
            the heuristics to help find good initial values for other model parameters
            for which none have been provided by the user. The value of `fixed_to` must
            lie within the bounds of the parameter.
        user_estimate: if not `None` and the parameter is not fixed, this value is
            used as an initial value during fitting rather than obtaining a value from
            the heuristics. This value may additionally be used by the heuristics to
            help find good initial values for other model parameters for which none
            have been provided by the user. The value of `user_estimate` must lie
            within the bounds of the parameter.
        heuristic: if both of `fixed_to` and `user_estimate` are `None`, this value is
            used as an initial value during fitting. It is set by the
            `estimate_parameters` method of the model in which the parameter is used
            and should not be set by the user.
    """

    scale_func: Callable[[float, float], float]
    lower_bound: float = -np.inf
    upper_bound: float = np.inf
    fixed_to: Optional[float] = None
    user_estimate: Optional[float] = None
    heuristic: Optional[float] = None
    scale_factor: Optional[float] = dataclasses.field(init=False, default=None)

    def __getattribute__(self, name):
        attr = object.__getattribute__(self, name)
        if name == "scale_factor":
            return attr

        scale_factor = self.scale_factor
        if attr is None or scale_factor is None:
            return attr

        if name in [
            "lower_bound",
            "upper_bound",
            "fixed_to",
            "user_estimate",
            "heuristic",
        ]:
            attr /= scale_factor

        return attr

    def __setattr__(self, name, value):
        scale_factor = self.scale_factor

        if None not in [scale_factor, value] and name in [
            "lower_bound",
            "upper_bound",
            "fixed_to",
            "user_estimate",
            "heuristic",
        ]:
            value *= scale_factor
        object.__setattr__(self, name, value)

    def rescale(self, x_scale: float, y_scale: float):
        """Rescales the parameter metadata based on the specified x and y data scale
        factors.
        """
        if self.scale_factor is not None:
            raise RuntimeError("Attempt to rescale an already rescaled model parameter")
        self.scale_factor = self.scale_func(x_scale, y_scale)

    def unscale(self):
        """Disables rescaling of the parameter metadata"""
        if self.scale_factor is None:
            raise RuntimeError(
                "Attempt to unscale a model parameter which was not rescaled."
            )
        self.scale_factor = None

    def get_initial_value(self, default: Optional[float] = None) -> float:
        """
        Get initial value.

        For fixed parameters, this is the value the parameter is fixed to. For floated
        parameters, it is the value used to seed the fit. In the latter case, the
        initial value is retrieved from `user_estimate` if that attribute is not
        `None`, otherwise `heuristic` is used.

        :param default: optional value to use if no other value is available
        """
        if self.fixed_to is not None:
            value = self.fixed_to
            if self.user_estimate is not None:
                raise ValueError(
                    "User estimates must not be provided for fixed parameters"
                )
        elif self.user_estimate is not None:
            value = self.user_estimate
        elif self.heuristic is not None:
            value = self.clip(self.heuristic)
        elif default is not None:
            value = self.clip(default)
        else:
            raise ValueError("No initial value specified")

        if value < self.lower_bound or value > self.upper_bound:
            raise ValueError("Initial value outside bounds.")

        return value

    def has_initial_value(self) -> bool:
        """
        Returns True if the parameter is fixed, has a user estimate or a heuristic.
        """
        values = [self.fixed_to, self.user_estimate, self.heuristic]
        return any([None is not value for value in values])

    def has_user_initial_value(self) -> bool:
        """Returns True if the parameter is fixed or has a user estimate"""
        return self.fixed_to is not None or self.user_estimate is not None

    def clip(self, value: float) -> float:
        """Clip value to lie between lower and upper bounds."""
        return np.clip(value, self.lower_bound, self.upper_bound)


class Model:
    """Base class for fit models.

    A model groups a function to be fitted with associated metadata (parameter names,
    default bounds etc) and heuristics. It is agnostic about the method of fitting or
    the data statistics.
    """

    def __init__(
        self,
        parameters: Optional[Dict[str, ModelParameter]] = None,
        internal_parameters: Optional[List[ModelParameter]] = None,
    ):
        """
        :param parameters: optional dictionary mapping names of model parameters to
            their metadata. This should be `None` (default) if the model has a static
            set of parameters in which case the parameter dictionary is generated from
            the call signature of :meth _func:. The model parameters are stored as
            `self.parameters` and may be modified after construction to change the model
            behaviour during fitting (e.g. to change the bounds, fixed parameters, etc).
        :param internal_parameters: optional list of "internal" model parameters, which
            are not exposed to the user as arguments of :meth func:. Internal parameters
            are rescaled in the same way as regular model parameters, but are not
            otherwise used by :class Model:. These are typically used by models which
            encapsulate / modify the behaviour of other models.
        """
        if parameters is None:
            spec = inspect.getfullargspec(self._func)
            for name in spec.args[2:]:
                assert isinstance(
                    spec.annotations[name], ModelParameter
                ), "Model parameters must be instances of `ModelParameter`"
            self.parameters = {
                name: copy.deepcopy(spec.annotations[name]) for name in spec.args[2:]
            }
        else:
            self.parameters = parameters
        self.internal_parameters = internal_parameters or []

    def __call__(self, x: TX, **kwargs: float) -> TY:
        """Evaluates the model.

        - keyword arguments specify values for model parameters (see model definition)
        - all model parameters which are not `fixed_to` a value by default must be
          specified
        - any parameters which are not specified default to their `fixed_to` values
        """
        args = {
            param_name: param_data.fixed_to
            for param_name, param_data in self.parameters.items()
            if param_data.fixed_to is not None
        }
        args.update(kwargs)
        return self.func(x, args)

    def can_rescale(self) -> Tuple[bool, bool]:
        """
        Returns a Tuple of bools specifying whether the model can be rescaled along
        the x- and y-axes
        """
        raise NotImplementedError

    def rescale(self, x_scale: float, y_scale: float):
        """Rescales the model parameters based on the specified x and y data scale
        factors.

        :param x_scale: x-axis scale factor
        :param y_scale: y-axis scale factor
        """
        for param_name, param_data in self.parameters.items():
            if param_data.scale_func == scale_undefined:
                raise RuntimeError(
                    f"Parameter {param_name} has an undefined scale function"
                )
            param_data.rescale(x_scale, y_scale)

        for param_data in self.internal_parameters:
            param_data.rescale(x_scale, y_scale)

    def unscale(self):
        """Disables rescaling of the model parameters."""
        parameters = list(self.parameters.values()) + self.internal_parameters
        for param_data in parameters:
            param_data.unscale()

    def get_num_y_channels(self) -> int:
        """Returns the number of y-axis dimensions ("channels") the model has."""
        raise NotImplementedError

    def clear_heuristics(self):
        """Clear the heuristics for all model parameters.

        This is mainly used in container-type models, where the parameter estimator
        my be run multiple times for the same model instance.
        """
        for param_data in self.parameters.values():
            param_data.heuristic = None
        for param_data in self.internal_parameters:
            param_data = None

    def get_num_x_axes(self) -> int:
        """Returns the number of x-axis dimensions the model has."""
        raise NotImplementedError

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result.

        To use the model as a function outside of a fit, :meth __call__: generally
        provides a more convenient interface.

        Overload this to provide a model function with a dynamic set of parameters,
        otherwise prefer to override `_func`.

        :param x: x-axis data
        :param param_values: dictionary of parameter values
        :returns: array of model values
        """
        return self._func(x, **param_values)

    def _func(
        self,
        x: TX,
    ) -> TY:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result.

        Overload this in preference to `func` unless the FitModel takes a
        dynamic set of parameters. Use : class ModelParameter: objects as the annotation
        for the parameters arguments. e.g.:

        ```
        def _func(self, x, a: ModelParameter(), b: ModelParameter()):
        ```

        A dictionary of `ModelParameter`s may be accessed via the `self.parameters`
        attribute. These parameters may be modified by the user to change the model
        behaviour during fitting (e.g. to change the bounds, fixed parameters, etc).

        :param x: x-axis data
        :returns: array of model values
        """
        raise NotImplementedError

    def estimate_parameters(
        self,
        x: TX,
        y: TY,
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If the model allows rescaling then rescaled
        units will be used everywhere (`x` and `y` as well as parameter values).

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        """
        raise NotImplementedError

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        return {}, {}


class Fitter:
    """Base class for fitters.

    Fitters perform maximum likelihood parameter estimation on a dataset under the
    assumption of a certain model and statistics (normal, binomial, etc) and store the
    results as attributes.

    Attributes:
        x: 1D ndarray of shape (num_samples,) containing x-axis values of valid points.
        y: 2D ndarray of shape (num_y_channels, num_samples) containing y-axis values
            of valid points.
        sigma: 2D ndarray of shape (num_y_channels, num_samples) containing y-axis
            standard errors where available.
        values: dictionary mapping model parameter names to their fitted values
        uncertainties: dictionary mapping model parameter names to their fit
            uncertainties. For sufficiently large datasets, well-formed problems and
            ignoring covariances these are the 1-sigma confidence intervals (roughly:
            there is a 1/3 chance that the real parameter values differ from their
            fitted values by more than this much)
        derived_values: dictionary mapping names of derived parameters (parameters which
            are not part of the fit, but are calculated by the model from the fitted
            parameter values) to their values
        derived_uncertainties: dictionary mapping names of derived parameters to their
            fit uncertainties
        initial_values: dictionary mapping model parameter names to the initial values
            used to seed the fit.
        model: the fit model
        free_parameters: list of names of the model parameters floated during the fit
        x_scale: the applied x-axis scale factor
        y_scale: the applied y-axis scale factor
    """

    x: TX
    y: TY
    sigma: Optional[TY]
    values: Dict[str, float]
    uncertainties: Dict[str, float]
    derived_values: Dict[str, float]
    derived_uncertainties: Dict[str, float]
    initial_values: Dict[str, float]
    model: Model
    free_parameters: List[str]
    x_scale: float
    y_scale: float

    def __init__(
        self,
        x: TX,
        y: TY,
        model: Model,
    ):
        """Fits a model to a dataset and stores the results.

        :param x: x-axis data
        :param y: y-axis data
        :param model: the model function to fit to. The model's parameter dictionary is
            used to configure the fit (set parameter bounds etc). Modify this before
            fitting to change the fit behaviour from the model class' defaults. The
            model is (deep) copied and stored as an attribute.
        """
        self.model = copy.deepcopy(model)

        self.x = np.array(x, dtype=np.float64, copy=True)
        self.y = np.atleast_2d(np.array(y, dtype=np.float64, copy=True))

        self.sigma = self.calc_sigma()
        self.sigma = np.atleast_2d(self.sigma) if self.sigma is not None else None

        if self.x.ndim != 1:
            raise ValueError("x-axis data must be a 1D array.")

        if self.y.ndim != 2:
            raise ValueError("y-axis data must be a 1D or 2D array.")

        if self.x.shape[0] != self.y.shape[1]:
            raise ValueError(
                "Number of x-axis and y-axis samples must match "
                f"(got {self.x.shape} and {self.y.shape})."
            )

        if self.y.shape[0] != self.model.get_num_y_channels():
            raise ValueError(
                f"Expected {self.model.get_num_y_channels()} y channels, "
                f"got {self.y.shape[0]}."
            )

        if self.sigma is not None and self.sigma.shape != self.y.shape:
            raise ValueError(
                f"Shapes of sigma and y must match (got {self.sigma.shape} and "
                f"{self.y.shape})."
            )

        valid_x = np.isfinite(self.x)
        valid_y = np.all(np.isfinite(self.y), axis=0)
        valid_pts = np.logical_and(valid_x, valid_y).squeeze()

        (inds,) = np.where(valid_pts)
        sorted_inds = inds[np.argsort(self.x[valid_pts])]

        self.x = self.x[sorted_inds]
        self.y = self.y[:, sorted_inds]

        if self.sigma is not None:

            self.sigma = self.sigma[:, sorted_inds]
            if np.any(self.sigma == 0) or not np.all(np.isfinite(self.sigma)):
                raise RuntimeError(
                    "Dataset contains points with zero or infinite uncertainty."
                )

        # Rescale coordinates to improve numerics (optimizers need to do things like
        # calculate numerical derivatives which is easiest if x and y are O(1)).
        #
        # Currently we use a single scale factor for all y channels. This may change in
        # future
        rescale_x, rescale_y = self.model.can_rescale()
        self.x_scale = np.max(np.abs(self.x)) if rescale_x else 1.0
        self.y_scale = np.max(np.abs(self.y)) if rescale_y else 1.0

        # Corner-case if a y-channel has values that are all 0
        if self.y_scale == 0 or not np.isfinite(self.y_scale):
            self.y_scale = 1
            rescale_y = False

        self.model.rescale(self.x_scale, self.y_scale)

        x_scaled = self.x / self.x_scale
        y_scaled = self.y / self.y_scale

        self.model.estimate_parameters(x_scaled, y_scaled)

        for param, param_data in self.model.parameters.items():
            if not param_data.has_initial_value():
                raise RuntimeError(
                    "No fixed_to, user_estimate or heuristic specified"
                    f" for parameter `{param}`."
                )

        self.fixed_parameters = {
            param_name: param_data.fixed_to
            for param_name, param_data in self.model.parameters.items()
            if param_data.fixed_to is not None
        }
        self.free_parameters = [
            param_name
            for param_name, param_data in self.model.parameters.items()
            if param_data.fixed_to is None
        ]

        if self.free_parameters == []:
            raise ValueError("Attempt to fit with no free parameters.")

        def free_func(
            x: Array[("num_samples",), np.float64], *free_param_values: float
        ) -> Array[("num_y_channels", "num_samples"), np.float64]:
            """Call the model function with the values of the free parameters."""
            params = {
                param: value
                for param, value in zip(self.free_parameters, list(free_param_values))
            }
            params.update(self.fixed_parameters)
            y = self.model.func(x, params)
            return y

        fitted_params, uncertainties = self._fit(
            x_scaled, y_scaled, self.model.parameters, free_func
        )
        fitted_params.update(
            {param: value for param, value in self.fixed_parameters.items()}
        )
        uncertainties.update({param: 0 for param in self.fixed_parameters.keys()})

        self.values = {
            param: value * self.model.parameters[param].scale_factor
            for param, value in fitted_params.items()
        }
        self.uncertainties = {
            param: value * self.model.parameters[param].scale_factor
            for param, value in uncertainties.items()
        }

        self.model.unscale()

        derived = self.model.calculate_derived_params(
            self.x, self.y, self.values, self.uncertainties
        )
        self.derived_values, self.derived_uncertainties = derived

        self.initial_values = {
            param: param_data.get_initial_value()
            for param, param_data in self.model.parameters.items()
        }

    def _fit(
        self,
        x: TX,
        y: TY,
        parameters: Dict[str, ModelParameter],
        free_func: Callable[..., TY],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Implementation of the parameter estimation.

        `Fitter` implementations must override this method to provide a fit with
        appropriate statistics.

        :param x: rescaled x-axis data, must be a 1D array
        :param y: rescaled y-axis data
        :param parameters: dictionary of rescaled model parameters
        :param free_func: convenience wrapper for the model function, taking only values
            for the fit's free parameters

        :returns: tuple of dictionaries mapping model parameter names to their fitted
            values and uncertainties.
        """
        raise NotImplementedError

    def evaluate(
        self,
        transpose_and_squeeze=False,
        x_fit: Optional[TX] = None,
    ) -> Tuple[TX, TY]:
        """Evaluates the model function using the fitted parameter set.

        :param transpose_and_squeeze: if True, array `y_fit` is transposed
            and squeezed before being returned. This is intended to be used
            for plotting, since matplotlib requires different y-series to be
            stored as columns.
        :param x_fit: optional x-axis points to evaluate the model at. If
            `None`, we use the values stored as attribute `x` of the fitter.

        :returns: tuple of x-axis values used and corresponding y-axis values
            of the fitted model
        """
        x_fit = x_fit if x_fit is not None else self.x
        y_fit = self.model.func(x_fit, self.values)

        if transpose_and_squeeze:
            return x_fit, y_fit.T.squeeze()
        return x_fit, y_fit

    def residuals(self) -> TY:
        """Returns an array of fit residuals."""
        return self.y - self.evaluate()[1]

    def calc_sigma(self) -> Optional[TY]:
        """Return an array of standard error values for each y-axis data point."""
        raise NotImplementedError



from typing import Dict, Tuple, TYPE_CHECKING

import numpy as np

from .. import common, Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


def _p_scale_func(x_scale: float, y_scale: float) -> float:
    if x_scale != 1.0:
        raise RuntimeError("Benchmarking model cannot be rescaled along x")
    return 1


class Benchmarking(Model):
    """Benchmarking success probability decay model

    y = (y0 - y_inf)*p^x + y_inf
    for sequence length x.

    Fit parameters (all floated by default unless stated otherwise):
      - p: depolarisation parameter
      - y0: SPAM fidelity estimate
      - y_inf: depolarisation offset (y-axis asymptote) (fixed to 1/2^n by default)

    Derived parameters:
      - e: error per Clifford = (1 - p) / alpha_n where alpha_n = 2^n / (2^n - 1)
      - e_spam: estimated SPAM error = 1 - y0
    """

    def __init__(self, num_qubits):
        """Init

        :param num_qubits: The number of qubits involved in the benchmarking sequence.
        """
        super().__init__()
        self.parameters["y_inf"].fixed_to = 1 / 2**num_qubits
        self.num_qubits = num_qubits
        self.alpha = 2**num_qubits / (2**num_qubits - 1)

    def get_num_y_channels(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[bool, bool]:
        return False, True

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        p: ModelParameter(lower_bound=0.0, upper_bound=1.0, scale_func=_p_scale_func),
        y0: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=common.scale_y,
        ),
        y_inf: ModelParameter(
            # fixed_to set to `1 / 2**num_qubits` in the constructor
            lower_bound=0,
            upper_bound=1,
            scale_func=common.scale_y,
        ),
    ) -> Array[("num_samples",), np.float64]:
        y = (y0 - y_inf) * p**x + y_inf
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
    ):
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        self.parameters["p"].heuristic = 1.0
        self.parameters["y0"].heuristic = max(y)
        self.parameters["y_inf"].heuristic = 1 / 2**self.num_qubits

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        p = fitted_params["p"]
        y0 = fitted_params["y0"]

        e = (1 - p) / self.alpha
        e_spam = 1 - y0

        p_err = fit_uncertainties["p"]
        y0_err = fit_uncertainties["y0"]

        derived_params = {}
        derived_params["e"] = e
        derived_params["e_spam"] = e_spam

        derived_uncertainties = {}
        derived_uncertainties["e"] = p_err / self.alpha
        derived_uncertainties["e_spam"] = y0_err

        return derived_params, derived_uncertainties



from .containers import AggregateModel, RepeatedModel
from .benchmarking import Benchmarking
from .exponential import Exponential
from .gaussian import Gaussian
from . import heuristics
from .laser_rabi import (
    LaserFlopFreqCoherent,
    LaserFlopFreqSqueezed,
    LaserFlopFreqThermal,
    LaserFlopTimeCoherent,
    LaserFlopTimeSqueezed,
    LaserFlopTimeThermal,
)
from .lorentzian import Lorentzian
from .molmer_sorensen import MolmerSorensenFreq, MolmerSorensenTime
from .polynomial import Power, Polynomial, Line, Parabola
from . import quantum_phys
from .rabi import RabiFlopFreq, RabiFlopTime
from .ramsey import Ramsey
from .rectangle import Rectangle
from .sigmoid import LogisticFunction
from .sinc import Sinc, Sinc2
from .sinusoid import Sinusoid
from .triangle import Triangle
from . import utils

from .cone import ConeSlice  # Relies on the triangle fit for parameter estimation



from typing import Dict, List, Optional, Tuple, TYPE_CHECKING

import numpy as np

from .utils import param_like
from .. import Model, ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float
    num_y_channels = float


class AggregateModel(Model):
    """Model formed by aggregating one or more models.

    When aggregating a number of identical models, use a :class RepeatedModel: instead.

    Aggregate models have a number of uses. For example:
      - joint fits to multiple data sets (pass the datasets in as y channels and use
        "common" parameters for any parameters which are fit jointly to all y channels).
      - fit multiple datasets simultaneously. This is useful, for example in automated
      tooling such as `ndscan`'s `OnlineAnalysis`.
    """

    def __init__(
        self,
        models: Dict[str, Model],
        common_params: Optional[
            Dict[str, Tuple[ModelParameter, List[Tuple[str, str]]]]
        ] = None,
    ):
        """
        :param models: The models to be aggregated. This should be a dictionary mapping
          model names to model instances. The model names are used as suffixes for names
          of model parameters and derived results. For example, if one of the aggregated
          models named `model` has a parameter `param`, the aggregate model will have a
          parameter `param_model`.

          The passed-in models are considered "owned" by the AggregateModel and should
          not be used / modified elsewhere.

        :param common_params: Optional dictionary specifying "common" model parameters.
          This feature allows multiple parameters (which can be from the same or
          different models) to be fit jointly to a single value. The common parameters
          are demoted to being `internal` model parameters and a new parameter is
          introduced to expose the common value to the user.

          The parameter metadata (limits, fixed_to, user_estimate, etc.) from the new
          parameter replaces the metadata for all parameters bound to it. Metadata set
          on the bound parameters is disregarded.

          The dictionary keys are the names of the new model parameters.

          The dictionary values are a tuple containing the new model template parameter
          and a list of parameters to bind to the new parameter. The bound parameter
          lists should be lists of tuples, specifying the parameters to bind to the new
          parameter. The tuples should contain two strings, specifying the name of the
          model which owns the common parameter, and the name of the model parameter to
          make common.

          The new model parameters inherit their metadata (limits etc.) from the
          template parameters, which are (deep) copied and are not modified.

        At present this class only supports models with a single y channel. This is just
        because no one got around to implementing it yet rather than any fundamental
        difficulty.
        """
        self.models = models
        common_params = common_params or {}

        # aggregate internal parameters from all models
        internal_parameters = []
        for model in self.models.values():
            internal_parameters += model.internal_parameters

        # organise the common parameter mapping data in ways that will be useful later

        # {new_param_name: [(model_name, param_name)]}
        self.common_param_list = {
            new_param_name: [] for new_param_name in common_params.keys()
        }

        # {model_name: [model_common_params]}
        self.model_common_params: Dict[str, List[str]] = {
            model_name: [] for model_name in self.models.keys()
        }

        # {(model_name, param_name): new_param_name}
        self.common_param_map: Dict[Tuple[str, str], str] = {}

        # {new_param_name: new_param}
        new_parameters: Dict[str, ModelParameter] = {}

        for new_param_name, (template_param, bound_params) in common_params.items():
            new_parameters[new_param_name] = param_like(template_param)
            for bind in bound_params:
                bound_model_name, bound_param_name = bind

                if bound_model_name not in self.models.keys():
                    raise ValueError(
                        f"Bound model name {bound_model_name} does not match any "
                        "aggregated model"
                    )

                model_parameters = self.models[bound_model_name].parameters
                if bound_param_name not in model_parameters.keys():
                    raise ValueError(
                        f"Bound parameter name {bound_param_name} does not match any "
                        "aggregated model"
                    )

                self.common_param_list[new_param_name] = [bind]
                self.model_common_params[bound_model_name].append(bound_param_name)
                self.common_param_map[bind] = new_param_name

        # aggregate non-common parameters from all models
        parameters: Dict[str, ModelParameter] = {}
        for model_name, model in self.models.items():
            if model.get_num_y_channels() != 1:
                raise ValueError(
                    "AggregateModel currently only supports models with a single y "
                    "channel"
                )
            parameters.update(
                {
                    f"{param_name}_{model_name}": param_data
                    for param_name, param_data in model.parameters.items()
                    if param_name not in self.model_common_params[model_name]
                }
            )

        duplicates = set(new_parameters.keys()).intersection(parameters.keys())
        if duplicates:
            raise ValueError(
                "New parameter names duplicate names of existing model parameters: "
                f"{duplicates}"
            )

        parameters.update(new_parameters)

        super().__init__(parameters=parameters, internal_parameters=internal_parameters)

    def get_num_y_channels(self) -> int:
        return len(self.models)

    def can_rescale(self) -> Tuple[bool, bool]:
        rescale_x, rescale_y = zip(
            *[model.can_rescale() for model in self.models.values()]
        )
        return all(rescale_x), all(rescale_y)

    def func(
        self,
        x: Array[("num_samples",), np.float64],
        param_values: Dict[str, float],
    ) -> Array[("num_y_channels", "num_samples"), np.float64]:
        ys = np.zeros((self.get_num_y_channels(), len(x)), dtype=np.float64)
        for idx, (model_name, model) in enumerate(self.models.items()):
            model_common_params = self.model_common_params[model_name]
            model_params = {
                param_name: param_values[f"{param_name}_{model_name}"]
                for param_name in model.parameters.keys()
                if param_name not in model_common_params
            }
            model_params.update(
                {
                    bound_param_name: param_values[
                        self.common_param_map[(model_name, bound_param_name)]
                    ]
                    for bound_param_name in model_common_params
                }
            )

            ys[idx, :] = np.atleast_2d(model.func(x, model_params))

        return ys

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
    ):
        for idx, (model_name, model) in enumerate(self.models.items()):
            # replace bound model parameters with new ones based on our template
            # NB we don't do this in __init__ because we want to capture subsequent
            # changes to parameter metadata
            for bound_param_name in self.model_common_params[model_name]:
                new_param_name = self.common_param_map[(model_name, bound_param_name)]
                new_param = self.parameters[new_param_name]
                model.parameters[bound_param_name] = param_like(new_param)

            model.estimate_parameters(x, y[idx])

        # use the mean value from all models as our heuristic for common params
        for new_param_name, binds in self.common_param_list.items():
            estimates = [
                self.models[model_name].parameters[param_name].get_initial_value()
                for model_name, param_name in binds
            ]
            self.parameters[new_param_name].heuristic = np.mean(estimates)

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_uncertainties = {}

        for idx, (model_name, model) in enumerate(self.models.items()):
            model_common_params = self.model_common_params[model_name]
            model_fitted_params = {
                param_name: fitted_params[f"{param_name}_{model_name}"]
                for param_name in model.parameters.keys()
                if param_name not in model_common_params
            }
            model_fit_uncertainties = {
                param_name: fit_uncertainties[f"{param_name}_{model_name}"]
                for param_name in model.parameters.keys()
                if param_name not in model_common_params
            }

            model_fitted_params.update(
                {
                    bound_param_name: fitted_params[
                        self.common_param_map[(model_name, bound_param_name)]
                    ]
                    for bound_param_name in model_common_params
                }
            )
            model_fit_uncertainties.update(
                {
                    bound_param_name: fit_uncertainties[
                        self.common_param_map[(model_name, bound_param_name)]
                    ]
                    for bound_param_name in model_common_params
                }
            )

            derived = model.calculate_derived_params(
                x=x,
                y=y[idx, :],
                fitted_params=model_fitted_params,
                fit_uncertainties=model_fit_uncertainties,
            )
            model_derived_params, model_derived_uncertainties = derived
            derived_params.update(
                {
                    f"{param_name}_{model_name}": value
                    for param_name, value in model_derived_params.items()
                }
            )
            derived_uncertainties.update(
                {
                    f"{param_name}_{model_name}": value
                    for param_name, value in model_derived_uncertainties.items()
                }
            )

        return derived_params, derived_uncertainties


class RepeatedModel(Model):
    """Model formed by repeating a `Model` one more times.

    The `RepeatedModel` has multiple y-channels, corresponding to the repetitions of
    the repeated model.

    Repeated models allow multiple datasets to be analysed simultaneously. This is
    useful, for example, when doing joint fits to datasets (using common parameters)
    or in automated tooling (for example ndscan OnlineAnalyses) which needs a single
    model for an entire dataset.
    """

    def __init__(
        self,
        model: Model,
        common_params: Optional[List[str]] = None,
        num_repetitions: int = 2,
        aggregate_results=False,
    ):
        """
        :param model: The repeated model, the implementation of `model` will be used to
          generate data for all y channels. This model is considered owned by the
          RepeatedModel and should not be used / modified elsewhere.
        :param common_params: optional list of names of model parameters, whose value is
          common to all y channels. All other model parameters are allowed to vary
          independently between the y channels
        :param num_repetitions: the number of times the model is repeated
        :param aggregate_results: determines whether derived results are aggregated or
          not (see below). The default behaviour is to not aggregate results. This is
          generally suitable when one wants access to the values of non-common
          parameters from the various repetitions. Aggregating results can be useful,
          for example, when all parameters are common across the repetitions and one
          wants a single set of values reported.

        The `RepeatedModel` has the following parameters and fit results:
          - all common parameters of the repeated model are parameters of the new model
          - for each independent (not common) parameter of the repeated model, `param`,
            the `RepeatedModel` has parameters `param_{n}` for n in
            [0, .., num_repitions-1]
          - for each independent parameter, `param`, the `RepeatedModel` model
            has additional fit results representing statistics between the repetitions.
            For each independent parameter `param`, the results dictionary will have
            additional quantities: `param_mean` and `param_peak_peak`.

        If :param aggregate_results: is `False` the rules for fit results follow those
        described previously. This is the default behaviour.

        If `aggregate_results` is `True`:
          - no statistical quantities are calculated for fit parameters or derived
            results
          - all derived results whose values and uncertainties are the same for all
            repetitions are aggregated together to give a single result. This has the
            same name as the original model (no suffix).
          - all derived results whose value is not the same for al repetitions are
            omitted.
        """
        self.aggregate_results = aggregate_results
        model_params = set(model.parameters.keys())
        common_params = set(common_params or [])

        if not common_params.issubset(model_params):
            raise ValueError(
                "Common parameters must be a subset of the model's parameters"
            )

        independent_params = set(model_params) - common_params
        params = {param: model.parameters[param] for param in common_params}
        for param in independent_params:
            params.update(
                {
                    f"{param}_{idx}": param_like(model.parameters[param])
                    for idx in range(num_repetitions)
                }
            )

        self.model = model
        self.common_params = common_params
        self.independent_params = independent_params
        self.num_repetitions = num_repetitions

        super().__init__(
            parameters=params, internal_parameters=self.model.internal_parameters
        )

    def get_num_y_channels(self) -> int:
        return self.num_repetitions * self.model.get_num_y_channels()

    def can_rescale(self) -> Tuple[bool, bool]:
        return self.model.can_rescale()

    def func(
        self,
        x: Array[("num_samples",), np.float64],
        param_values: Dict[str, float],
    ) -> Array[("num_y_channels", "num_samples"), np.float64]:
        values = {param: param_values[param] for param in self.common_params}

        dim = self.model.get_num_y_channels()
        ys = np.zeros((self.get_num_y_channels(), len(x)))
        for idx in range(self.num_repetitions):
            values.update(
                {
                    param: param_values[f"{param}_{idx}"]
                    for param in self.independent_params
                }
            )
            ys[idx * dim : (idx + 1) * dim, :] = np.atleast_2d(
                self.model.func(x, values)
            )

        return ys

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
    ):
        dim = self.model.get_num_y_channels()

        common_heuristics = {
            param: np.zeros(self.num_repetitions + 1) for param in self.common_params
        }

        for idx in range(self.num_repetitions):
            params = {
                param: self.parameters[f"{param}_{idx}"]
                for param in self.independent_params
            }

            params.update(
                {param: self.parameters[param] for param in self.common_params}
            )

            # Reset heuristics before each iteration
            for param_data in params.values():
                param_data.heuristic = None

            self.model.parameters = params
            self.model.estimate_parameters(x, y[idx * dim : (idx + 1) * dim])

            for param in self.common_params:
                common_heuristics[param][idx] = params[param].get_initial_value()

        # Combine the heuristics for the repetitions to find the best set of common
        # parameter values
        for param in self.common_params:
            common_heuristics[param][-1] = np.mean(common_heuristics[param][:-1])

        param_estimates = {
            param_name: self.parameters[param_name].get_initial_value()
            for param_name in self.parameters.keys()
            if param_name not in self.common_params
        }
        costs = np.zeros(self.num_repetitions + 1)
        for idx in range(self.num_repetitions + 1):
            y_idx = self.__call__(
                x=x,
                **param_estimates,
                **{
                    param_name: common_heuristics[param_name][idx]
                    for param_name in self.common_params
                },
            )
            costs[idx] = np.sqrt(np.sum((y - y_idx) ** 2))
        best_heuristic = np.argmin(costs)

        for param in self.common_params:
            self.parameters[param].heuristic = common_heuristics[param][best_heuristic]

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_uncertainties = {}

        dim = self.model.get_num_y_channels()

        # {param_name: [derived_values_from_each_repetition]}
        derived_params_reps = None
        derived_uncertainties_reps = None
        derived_param_names = None

        common_fitted_params = {
            param: fitted_params[param] for param in self.common_params
        }
        common_fit_uncertainties = {
            param: fit_uncertainties[param] for param in self.common_params
        }

        for idx in range(self.num_repetitions):
            rep_params = {
                param: fitted_params[f"{param}_{idx}"]
                for param in self.independent_params
            }
            rep_params.update(common_fitted_params)

            rep_uncertainties = {
                param: fit_uncertainties[f"{param}_{idx}"]
                for param in self.independent_params
            }
            rep_uncertainties.update(common_fit_uncertainties)

            derived = self.model.calculate_derived_params(
                x=x,
                y=y[idx * dim : (idx + 1) * dim],
                fitted_params=rep_params,
                fit_uncertainties=rep_uncertainties,
            )
            rep_derived_params, rep_derived_uncertainties = derived

            if derived_params_reps is None:
                derived_param_names = rep_derived_params.keys()
                # We can't preallocate prior to this because we don't know what the
                # model's derived parameter values are until we've calculated them
                derived_params_reps = {
                    param_name: [0.0] * self.num_repetitions
                    for param_name in derived_param_names
                }
                derived_uncertainties_reps = {
                    param_name: [0.0] * self.num_repetitions
                    for param_name in derived_param_names
                }

            for param_name in derived_param_names:
                derived_params_reps[param_name][idx] = rep_derived_params[param_name]
                derived_uncertainties_reps[param_name][idx] = rep_derived_uncertainties[
                    param_name
                ]

        def add_statistics(values, uncertainties, param_names):
            for param_name in param_names:
                param_values = np.array(
                    [
                        values[f"{param_name}_{idx}"]
                        for idx in range(self.num_repetitions)
                    ]
                )
                param_uncerts = np.array(
                    [
                        uncertainties[f"{param_name}_{idx}"]
                        for idx in range(self.num_repetitions)
                    ]
                )

                derived_params[f"{param_name}_mean"] = np.mean(param_values)
                derived_uncertainties[f"{param_name}_mean"] = (
                    np.sqrt(np.sum(param_uncerts**2)) / self.num_repetitions
                )

                derived_params[f"{param_name}_peak_peak"] = np.ptp(param_values)
                derived_uncertainties[f"{param_name}_peak_peak"] = float("nan")

        if self.aggregate_results:
            keep = [
                param_name
                for param_name in derived_param_names
                if np.all(
                    np.isclose(
                        derived_params_reps[param_name],
                        derived_params_reps[param_name][0],
                    )
                )
                and np.all(
                    np.isclose(
                        derived_uncertainties_reps[param_name],
                        derived_uncertainties_reps[param_name][0],
                    )
                )
            ]
            derived_params.update(
                {param_name: derived_params_reps[param_name][0] for param_name in keep}
            )
            derived_uncertainties.update(
                {
                    param_name: derived_uncertainties_reps[param_name][0]
                    for param_name in keep
                }
            )
        else:
            for param_name in derived_param_names:
                derived_params.update(
                    {
                        f"{param_name}_{idx}": derived_params_reps[param_name][idx]
                        for idx in range(self.num_repetitions)
                    }
                )
                derived_uncertainties.update(
                    {
                        f"{param_name}_{idx}": derived_uncertainties_reps[param_name][
                            idx
                        ]
                        for idx in range(self.num_repetitions)
                    }
                )

            add_statistics(
                derived_params,
                derived_uncertainties,
                param_names=rep_derived_params.keys(),
            )
            add_statistics(
                fitted_params,
                fit_uncertainties,
                param_names=self.independent_params,
            )

        return derived_params, derived_uncertainties


class MappedModel(Model):
    """`Model` wrapping another `Model` with renamed parameters"""

    def __init__(
        self,
        model: Model,
        param_mapping: Dict[str, str],
        fixed_params: Optional[Dict[str, float]] = None,
        derived_result_mapping: Optional[Dict[str, Optional[str]]] = None,
    ):
        """Init

        :param model: The wrapped model. This model is considered "owned" by the
            MappedModel and should not be modified / used elsewhere.
        :param param_mapping: dictionary mapping names of parameters in the new
            model to names of parameters used in the wrapped model.
        :param fixed_params: dictionary mapping names of parameters used in the
            wrapped model to values they are fixed to in the new model. These
            will not be parameters of the new model.
        :param derived_result_mapping: optional dictionary mapping names of derived
            result in the new model to names of derived results in the wrapped model.
            Derived results may be renamed to `None` to exclude them from the model's
            outputs.
        """
        self.model = model
        self.derived_result_mapping = derived_result_mapping or {}
        wrapped_params = self.model.parameters

        fixed_params = fixed_params or {}

        if unknown_mapped_params := set(param_mapping.values()) - wrapped_params.keys():
            raise ValueError(
                "The target of parameter mappings must be parameters of the inner "
                f"model. The following mapping targets are not: {unknown_mapped_params}"
            )

        if unknown_fixed_params := fixed_params.keys() - wrapped_params.keys():
            raise ValueError(
                "Fixed parameters must be parameters of the inner model. The "
                f"follow fixed parameters are not: {unknown_fixed_params}"
            )

        if missing_params := wrapped_params.keys() - (
            fixed_params.keys() | param_mapping.values()
        ):
            raise ValueError(
                "All parameters of the inner model must be either mapped of "
                "fixed. The following inner model parameters are neither: "
                f"{missing_params}"
            )

        if duplicated_params := fixed_params.keys() & param_mapping.values():
            raise ValueError(
                "Parameters cannot be both mapped and fixed. The following "
                f"parameters are both: {duplicated_params}"
            )

        self.fixed_params = {
            param_name: self.model.parameters[param_name]
            for param_name in fixed_params.keys()
        }
        for param_name, fixed_to in fixed_params.items():
            self.fixed_params[param_name].fixed_to = fixed_to

        self.param_mapping = param_mapping
        exposed_params = {
            new_name: self.model.parameters[old_name]
            for new_name, old_name in param_mapping.items()
        }
        internal_parameters = (
            list(self.fixed_params.values()) + self.model.internal_parameters
        )

        super().__init__(
            parameters=exposed_params,
            internal_parameters=internal_parameters,
        )

    def get_num_y_channels(self) -> int:
        return self.model.get_num_y_channels()

    def can_rescale(self) -> Tuple[bool, bool]:
        return self.model.can_rescale()

    def func(
        self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
    ) -> Array[("num_samples", "num_y_channels"), np.float64]:
        new_params = {
            old_name: param_values[new_name]
            for new_name, old_name in self.param_mapping.items()
        }
        new_params.update(
            {
                param_name: param_data.fixed_to
                for param_name, param_data in self.fixed_params.items()
            }
        )
        return self.model.func(x, new_params)

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples", "num_y_channels"), np.float64],
    ):
        self.model.estimate_parameters(x, y)

        for param_name, param_data in self.parameters.items():
            old_name = self.param_mapping[param_name]
            old_param = self.model.parameters[old_name]
            param_data.heuristic = old_param.heuristic

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        new_fitted_params = {
            old_name: fitted_params[new_name]
            for new_name, old_name in self.param_mapping.items()
        }
        new_fitted_params.update(
            {
                param_name: param_data.fixed_to
                for param_name, param_data in self.fixed_params.items()
            }
        )

        new_fit_uncertainties = {
            old_name: fit_uncertainties[new_name]
            for new_name, old_name in self.param_mapping.items()
        }
        new_fit_uncertainties.update(
            {param_name: 0.0 for param_name, param_data in self.fixed_params.items()}
        )
        derived = self.model.calculate_derived_params(
            x=y,
            y=y,
            fitted_params=new_fitted_params,
            fit_uncertainties=new_fit_uncertainties,
        )
        mapping = self.derived_result_mapping
        for derived_dict in derived:
            for new_result_name, model_result_name in mapping.items():
                if new_result_name is None:
                    del derived_dict[model_result_name]
                    continue

                if model_result_name not in derived_dict:
                    raise ValueError(
                        f"Mapped derived result '{model_result_name}' not found."
                    )
                value = derived_dict.pop(model_result_name)
                if new_result_name in derived_dict:
                    raise ValueError(
                        f"Mapped derived result '{new_result_name}' duplicates "
                        "existing derived resut name."
                    )
                derived_dict[new_result_name] = value

        return derived



from .common import ModelParameter, Model, Fitter
from .MLE import MLEFitter

from .binomial import BinomialFitter
from .normal import NormalFitter
from . import models
from . import utils
from . import validators

from . import multi_x



import numpy as np
from typing import Dict, Tuple, TYPE_CHECKING

from . import heuristics
from .. import common, Model, ModelParameter
from ..utils import Array
from . import utils

if TYPE_CHECKING:
    num_samples = float


class Sinusoid(Model):
    """Generalised sinusoid fit according to:
    y = Gamma * a * sin[omega * (x - x0) + phi] + y0
    where Gamma = exp(-x / tau).

    Fit parameters (all floated by default unless stated otherwise):
      - a: initial (x = 0) amplitude of the sinusoid
      - omega: angular frequency
      - phi: phase offset (radians)
      - y0: y-axis offset
      - x0: x-axis offset (fixed to 0 by default)
      - tau: decay/growth constant (fixed to np.inf by default)

    Derived parameters:
      - f: frequency
      - phi_cosine: cosine phase (phi + pi/2)
      - contrast: peak-to-peak amplitude of the pure sinusoid
      - min/max: min / max values of the pure sinusoid
      - period: period of oscillation
      - TODO: peak values of the damped sinusoid as well as `x` value that the peak
          occurs at.

    All phases are in radians, frequencies are in angular units.

    x0 and phi0 are equivalent parametrisations for the phase offset, but in some cases
    it works out convenient to have access to both (e.g. one as a fixed offset, the
    other floated). At most one of them should be floated at once. By default, x0 is
    fixed at 0 and phi0 is floated.
    """

    def get_num_y_channels(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[bool, bool]:
        return True, True

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        a: ModelParameter(lower_bound=0, scale_func=common.scale_y),
        omega: ModelParameter(lower_bound=0, scale_func=common.scale_x_inv),
        phi: utils.PeriodicModelParameter(
            period=2 * np.pi,
            offset=-np.pi,
            scale_func=common.scale_invariant,
        ),
        y0: ModelParameter(scale_func=common.scale_y),
        x0: ModelParameter(fixed_to=0, scale_func=common.scale_x),
        tau: ModelParameter(
            lower_bound=0,
            fixed_to=np.inf,
            scale_func=common.scale_x,
        ),
    ) -> Array[("num_samples",), np.float64]:
        Gamma = np.exp(-x / tau)
        y = Gamma * a * np.sin(omega * (x - x0) + phi) + y0
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
    ):
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        # We don't have good heuristics for these parameters
        self.parameters["y0"].heuristic = np.mean(y)
        self.parameters["tau"].heuristic = np.max(x)

        omega, spectrum = heuristics.get_spectrum(
            x, y, density_units=False, trim_dc=True
        )
        spectrum = np.abs(spectrum)
        peak = np.argmax(spectrum)

        self.parameters["a"].heuristic = spectrum[peak]
        self.parameters["omega"].heuristic = omega[peak]

        phi, _ = heuristics.param_min_sqrs(
            model=self,
            x=x,
            y=y,
            scanned_param="phi",
            scanned_param_values=np.linspace(-np.pi, np.pi, num=20),
            defaults={"x0": 0},
        )

        phi = self.parameters["phi"].clip(phi)

        if self.parameters["x0"].fixed_to is None:
            if self.parameters["phi"].fixed_to is None:
                raise ValueError("Only one of 'x0' and 'phi' may be floated at once")

            self.parameters["phi"].heuristic = 0
            self.parameters["x0"].heuristic = (
                -phi / self.parameters["omega"].get_initial_value()
            )
        else:
            self.parameters["phi"].heuristic = phi
            self.parameters["x0"].heuristic = 0.0

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_params["f"] = fitted_params["omega"] / (2 * np.pi)
        derived_params["phi_cosine"] = fitted_params["phi"] + np.pi / 2
        derived_params["contrast"] = 2 * np.abs(fitted_params["a"])
        derived_params["min"] = fitted_params["y0"] - np.abs(fitted_params["a"])
        derived_params["max"] = fitted_params["y0"] + np.abs(fitted_params["a"])
        derived_params["period"] = 2 * np.pi / fitted_params["omega"]

        derived_uncertainties = {}
        derived_uncertainties["f"] = fit_uncertainties["omega"] / (2 * np.pi)
        derived_uncertainties["phi_cosine"] = fit_uncertainties["phi"]
        derived_uncertainties["contrast"] = 2 * fit_uncertainties["a"]
        derived_uncertainties["min"] = np.sqrt(
            fit_uncertainties["y0"] ** 2 + fit_uncertainties["a"] ** 2
        )
        derived_uncertainties["max"] = np.sqrt(
            fit_uncertainties["y0"] ** 2 + fit_uncertainties["a"] ** 2
        )
        derived_uncertainties["period"] = (
            2 * np.pi * fit_uncertainties["omega"] / (fitted_params["omega"] ** 2)
        )

        return derived_params, derived_uncertainties



from typing import TYPE_CHECKING

from .common import Model2D
from .. import models


if TYPE_CHECKING:
    num_samples_ax_0 = float
    num_samples_ax_1 = float
    num_y_channels = float


class Gaussian2D(Model2D):
    """2D Gaussian according to:
    ```
    z = (
        a / ((sigma * sqrt(2*pi)) * (sigma * sqrt(2*pi)))
        * exp(-0.5*((x-x0)/(sigma_x))^2 -0.5*((y-y0)/(sigma_y))^2) + z0
    ```
    Parameters are:
      - a
      - x0
      - y0
      - sigma_x
      - sigma_y
      - z0

    Derived results are:
      - FWHMH_x
      - FWHMH_y
      - w0_x
      - w0_y
      - peak_y

    See ionics_fits.models.Gaussian for details.
    """

    def __init__(self):
        # TODO: once we have proper support for 2D models, we should provide
        # non-empty model names and Wrap the 2D model rather than wrapping the 1D
        # models
        inner_model = models.containers.MappedModel(
            model=models.Gaussian(),
            param_mapping={"a_x": "a", "x0_x": "x0", "sigma_x": "sigma", "z0": "y0"},
            derived_result_mapping={"FWHMH_x": "FWHMH", "w0_x": "w0", None: "peak"},
        )
        outer_model = models.containers.MappedModel(
            model=models.Gaussian(),
            param_mapping={
                "a": "a",
                "x0_y": "x0",
                "sigma_y": "sigma",
            },
            fixed_params={"y0": 0},
            derived_result_mapping={"FWHMH_y": "FWHMH", "w0_y": "w0"},
        )

        super().__init__(
            models=(inner_model, outer_model),
            model_names=("", ""),
            result_params=("a_x",),
        )


class Parabola2D(Model2D):
    """2D Parabola according to:
    ```
    z = k_x * (x - x0)^2 + k_y *(y - y0) + z0
    ```

    Parameters are:
      - x0
      - y0
      - k_x
      - k_y
      - y0

    See ionics_fits.models.Parabola for details.
    """

    def __init__(self):
        # TODO: once we have proper support for 2D models, we should provide
        # non-empty model names and Wrap the 2D model rather than wrapping the 1D
        # models
        inner_model = models.containers.MappedModel(
            model=models.Parabola(),
            param_mapping={"x0": "x0", "y0_x": "y0", "k_x": "k"},
        )
        outer_model = models.containers.MappedModel(
            model=models.Parabola(),
            param_mapping={"y0": "x0", "z0": "y0", "k_y": "k"},
        )

        super().__init__(
            models=(inner_model, outer_model),
            model_names=("", ""),
            result_params=("y0_x",),
        )


class Cone2D(Model2D):
    """2D Cone Model.

    Parameters are (x params inherit from :class :class ionics_fits.models.ConeSlide:
    while y params inherit from :class ionics_fits.models.Triangle:
      - x0_x
      - x0_y
      - k_x
      - k_y
      - y0

     See ionics_fits.models.ConeSlice and ionics_fits.models.triangle for details.
    """

    def __init__(self):
        # TODO: once we have proper support for 2D models, we should provide
        # non-empty model names and Wrap the 2D model rather than wrapping the 1D
        # models
        inner_model = models.containers.MappedModel(
            model=models.ConeSlice(),
            param_mapping={"alpha": "alpha", "y0": "z0", "x0_x": "x0", "k_x": "k"},
        )

        triangle = models.Triangle()
        triangle.parameters["y0"].fixed_to = 0
        outer_model = models.containers.MappedModel(
            model=triangle,
            param_mapping={"x0_y": "x0", "k_y": "k"},
            fixed_params={
                param_name: param_data.fixed_to
                for param_name, param_data in triangle.parameters.items()
                if param_data.fixed_to is not None
            },
        )

        super().__init__(
            models=(inner_model, outer_model),
            model_names=("", ""),
            result_params=("alpha",),
        )



from typing import Tuple, TYPE_CHECKING
import numpy as np

from .. import common, Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


class LogisticFunction(Model):
    """Logistic function model according to:
    y = a / (1 + exp(-k*(x - x0))) + y0

    Fit parameters (all floated by default unless stated otherwise):
      - a: y-axis scale factor
      - y0: y-axis offset
      - x0: x-axis offset
      - k: logistic growth rate (steepness of the curve)

    Derived parameters:
      None
    """

    def get_num_y_channels(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[bool, bool]:
        return True, True

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        a: ModelParameter(scale_func=common.scale_y),
        y0: ModelParameter(scale_func=common.scale_y),
        x0: ModelParameter(scale_func=common.scale_x),
        k: ModelParameter(scale_func=common.scale_x_inv, lower_bound=0),
    ) -> Array[("num_samples",), np.float64]:
        return a / (1 + np.exp(-k * (x - x0))) + y0

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
    ):
        y = y.squeeze()
        sgn = +1 if y[-1] > y[0] else -1
        y0 = self.parameters["y0"].get_initial_value(y[0])
        self.parameters["y0"].heuristic = y0

        a = self.parameters["a"].get_initial_value(sgn * y.ptp())
        self.parameters["a"].heuristic = a

        y_centre = y0 + a / 2
        x0 = self.parameters["x0"].get_initial_value(x[np.argmin(np.abs(y - y_centre))])
        self.parameters["x0"].heuristic = x0

        FWHMH_pts = np.abs(y - y_centre) < np.abs(a / 4)
        x_FWHMH = x[FWHMH_pts]
        FWHMH = x_FWHMH.ptp()

        k = self.parameters["k"].get_initial_value(1.098 / (FWHMH / 2))
        self.parameters["k"].heuristic = k



from typing import Tuple
import numpy as np


import ionics_fits as fits


class FitValidator:
    """Base class for fit validators"""

    def window_fun(self, fit_results: fits.Fitter):
        """Returns a boolean array of x-axis points which should be included in
        fit validation. This implementation includes all x-axis data points;
        override to only include a subset.
        """
        return np.full(fit_results.x.shape, True)

    def validate(self, fit_results: fits.Fitter) -> Tuple[bool, float]:
        """Returns a Tuple of:

         bool: `True` if the fit was successful otherwise `False`.
         float: significance from the applied test as a number between 0.0 (
           complete failure) and 1.0 (perfect fit).

        Subclasses must override this method
        """
        raise NotImplementedError


class NSigmaValidator(FitValidator):
    def __init__(self, n_sigma: float = 3.0, significance_threshold: float = 0.75):
        """Fit validator which checks that at least
        :param significance_threshold: of points lie within :param n_sigma:
        standard errors of the fitted value.

        This is a relatively forgiving (easy to configure in a way that gives
        minimal "good" fits which fail validation) general-purpose fit validator.
        """
        self.n_sigma = n_sigma
        self.significance_threshold = significance_threshold

    def validate(self, fit_results: fits.Fitter) -> Tuple[bool, float]:
        if fit_results.sigma is None:
            raise ValueError(
                "Cannot validate fit without standard errors for each point"
            )

        x = fit_results.x
        y = fit_results.y
        sigma = fit_results.sigma

        if list(fit_results.x) == 1:
            window = self.window_fun(fit_results)

            x = fit_results.x[window]
            y = fit_results.y[:, window]
            sigma = fit_results.sigma[:, window]

        _, y_fit = fit_results.evaluate(x_fit=x)

        errs = np.abs((y - y_fit) / sigma)
        errs = np.sort(errs.flatten())

        good_points = errs < self.n_sigma
        good_fraction = sum(good_points) / len(good_points)

        return good_fraction >= self.significance_threshold, good_fraction



from typing import Dict, Tuple, TYPE_CHECKING
import numpy as np

from .. import common, Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


class Exponential(Model):
    """Exponential function according to:
    y(x < x_dead) = y0
    y(x >= x_dead) = y0 + (y_inf-y0)*(1-exp(-(x-x_dead)/tau))

    Fit parameters (all floated by default unless stated otherwise):
      - x_dead: x-axis "dead time" (fixed to 0 by default)
      - y0: initial (x = x_dead) y-axis offset
      - y_inf: y-axis asymptote (i.e. y(x - x_0 >> tau) => y_inf)
      - tau: decay constant

    Derived parameters:
      - x_1_e: x-axis value for 1/e decay including dead time (`x_1_e = x0 + tau`)
    """

    def get_num_y_channels(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[bool, bool]:
        return True, True

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x_dead: ModelParameter(
            lower_bound=0,
            fixed_to=0,
            scale_func=common.scale_x,
        ),
        y0: ModelParameter(scale_func=common.scale_y),
        y_inf: ModelParameter(scale_func=common.scale_y),
        tau: ModelParameter(lower_bound=0, scale_func=common.scale_x),
    ) -> Array[("num_samples",), np.float64]:
        y = y0 + (y_inf - y0) * (1 - np.exp(-(x - x_dead) / tau))
        y = np.where(x >= x_dead, y, y0)
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
    ):
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        # Exponentials are generally pretty easy to fit so we keep the estimator simple
        self.parameters["x_dead"].heuristic = 0
        self.parameters["y0"].heuristic = y[0]
        self.parameters["y_inf"].heuristic = y[-1]
        self.parameters["tau"].heuristic = x.ptp()

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {"x_1_e": fitted_params["x_dead"] + fitted_params["tau"]}
        derived_uncertainties = {
            "x_1_e": np.sqrt(
                fit_uncertainties["x_dead"] ** 2 + fit_uncertainties["tau"] ** 2
            )
        }
        return derived_params, derived_uncertainties



import copy
import dataclasses
import numpy as np
from typing import Any, Dict, Optional, Type, TYPE_CHECKING, TypeVar
from ..common import Model, ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float
    num_y_channels = float
    num_spectrum_samples = float

TModel = TypeVar("TModel", bound=Type[Model])


def param_like(
    template_param: ModelParameter, overrides: Optional[Dict[str, Any]] = None
) -> ModelParameter:
    """Returns a new parameter based on a template.

    :param template_param: the returned parameter is a (deep) copy of the template
      parameter.
    :param overrides: optional dictionary of attributes of the template parameter to
      replace.
    """
    new_param = copy.deepcopy(template_param)

    overrides = overrides or {}
    for attr, value in overrides.items():
        setattr(new_param, attr, value)

    return new_param


@dataclasses.dataclass
class PeriodicModelParameter(ModelParameter):
    period: float = 1
    offset: float = 0
    lower_bound: float = dataclasses.field(init=False)
    upper_bound: float = dataclasses.field(init=False)

    def __post_init__(self):
        self.lower_bound = 1.5 * self.offset
        self.upper_bound = 1.5 * (self.offset + self.period)

    upper_bound = 1.5 * np.pi

    def clip(self, value: float):
        """Clip value to lie between lower and upper bounds."""
        value = value - self.offset
        return (value % self.period) + self.offset


def rescale_model_x(model_class: TModel, x_scale: float) -> TModel:
    """Rescales the x-axis for a model class.

    This is commonly used to convert models between linear and angular units.

    :param model_class: model class to rescale
    :param x_scale: multiplicative x-axis scale factor. To convert a model that takes
      x in angular units and convert to one that takes x in linear units use
      `x_scale = 2 * np.pi`
    """

    class ScaledModel(model_class):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__x_scale = x_scale
            self.__rescale = True

        def func(
            self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
        ) -> Array[("num_samples", "num_y_channels"), np.float64]:
            x = (x * self.__x_scale) if self.__rescale else x
            return super().func(x, param_values)

        def estimate_parameters(
            self,
            x: Array[("num_samples",), np.float64],
            y: Array[("num_samples", "num_y_channels"), np.float64],
        ):
            # avoid double rescaling if estimate_parameters calls self.func internally
            self.__rescale = False
            super().estimate_parameters(x * self.__x_scale, y)
            self.__rescale = True

    ScaledModel.__name__ = model_class.__name__
    ScaledModel.__doc__ = model_class.__doc__

    return ScaledModel



import copy
import logging
from typing import Dict, List, Optional, Tuple, Union, TYPE_CHECKING

import numpy as np


from .. import Fitter, Model, NormalFitter
from ..models import RepeatedModel
from ..utils import Array, ArrayLike


if TYPE_CHECKING:
    num_free_params = float
    num_samples = float
    num_samples_x_flat = float
    num_samples_ax_0 = float
    num_samples_ax_1 = float
    num_values = float
    num_y_channels = float
    num_samples_flattened = float


TX2D = Tuple[
    ArrayLike[("num_samples_ax_0",), np.float64],
    ArrayLike[("num_samples_ax_1",), np.float64],
]
# NB ordering of x-axis samples here is chosen to play nicely with np.meshgrid
TY2D = Union[
    ArrayLike[("num_y_channels", "num_samples_ax_0", "num_samples_ax_1"), np.float64],
    ArrayLike[("num_samples_ax_0", "num_samples_ax_1"), np.float64],
]

logger = logging.getLogger(__name__)


class Model2D(Model):
    """Combines a pair of :class Model:s, which are each a function of 1 x-axis, to
    make a new :class Model:, which is a function of 2 x-axes.

    All y-axis data is generated from the output of the first model; the output from
    the second model provides the values of certain "result" parameters used by the
    first model. In other words:
      ```
      model_0 = models[0] = f(x_axis_0)
      model_1 = models[1] = g(x_axis_1)
      y(x_0, x_1) = model_0(x_0 | result_params = model_1(x_1))
      ```

    An intrinsic limitation of this approach is that the 2D fit function must be
    separable into functions of the two axes. This means, for example, that the 1D
    fit-functions must be aligned with the  two x axes. For example, one can't represent
    a non axis-aligned Gaussian in this way.

    Model parameters and results:
    - All parameters from the two models - apart from the first model's *result
      parameters* - are parameters of the 2D model.
    - All derived results from the two models are included in the :class Model2D:'s
      derived results.
    - A parameter/derived result named `param` from a model named `model` is
      exposed as a parameter / result of the :class Model2D: named `param_model`.
    - A :class WrappedModel: can be used to provide custom naming schemes
    """

    def __init__(
        self,
        models: Tuple[Model, Model],
        result_params: Tuple[str],
        model_names: Optional[Tuple[str, str]] = None,
    ):
        """
        :param models: Tuple containing the two :class Model:s to be combined to make
            the 2D model. The model instances are considered "owned" by the 2D model
            (they are not copied). They should not be referenced externally.
        :param result_params: tuple of names of "result parameters" of the first model,
            whose value is found by evaluating the second model. The order of parameter
            names in this tuple must match the order of result channels for the second
            model.
        :param model_names: optional tuple of names for the two models. These are used
            when generating names for fit results and derived parameters. Empty strings
            are allowed so long as the two models do not share any parameter names. If
            this argument is `None` the model names default to `x0` and `x1`
            respectively. If a model name is an empty string, the trailing underscore
            is omitted in parameter / result names.
        """
        self.models = models
        self.result_params = result_params
        self.model_names = model_names if model_names is not None else ("x0", "x1")

        missing = set(self.result_params) - set(self.models[0].parameters.keys())
        if missing:
            raise ValueError(
                "Result parameters must be parameters of the first model. Unrecognised "
                f"result parameter names are: {missing}"
            )

        if len(self.result_params) != self.models[1].get_num_y_channels():
            raise ValueError(
                f"{len(self.result_params)} result parameters passed when second model "
                f"requires {self.models[1].get_num_y_channels()}"
            )

        # Map the parameters of the 1D models onto parameters of the 2D models
        parameters = {}
        internal_parameters = [
            self.models[0].parameters[param_name] for param_name in self.result_params
        ]

        self.model_param_maps = ({}, {})  # model_param_name: new_param_name

        for model_idx, model_name in enumerate(self.model_names):
            # Result parameters are internal parameters of the 2D model
            exclusions = self.result_params if model_idx == 0 else []
            suffix = f"_{model_name}" if model_name else ""
            self.model_param_maps[model_idx].update(
                {
                    model_param_name: f"{model_param_name}{suffix}"
                    for model_param_name in self.models[model_idx].parameters.keys()
                    if model_param_name not in exclusions
                }
            )

        duplicates = set(self.model_param_maps[0].values()).intersection(
            set(self.model_param_maps[1].values())
        )
        if duplicates:
            raise ValueError(
                f"Duplicate parameter names found between the two models: {duplicates}."
            )

        # Create the parameters / internal parameters for the 2D model
        for model_idx, param_map in enumerate(self.model_param_maps):
            model = self.models[model_idx]
            parameters.update(
                {
                    new_param_name: model.parameters[model_param_name]
                    for model_param_name, new_param_name in param_map.items()
                }
            )
            internal_parameters += model.internal_parameters

        super().__init__(parameters=parameters, internal_parameters=internal_parameters)

    def can_rescale(self) -> Tuple[bool, bool]:
        raise NotImplementedError

    def get_num_y_channels(self) -> int:
        return self.models[0].get_num_y_channels()

    def get_num_x_axes(self) -> int:
        return 2

    def func(self, x: TX2D, param_values: Dict[str, float]) -> TY2D:
        x_ax_0, x_ax_1 = [np.array(x_ax) for x_ax in x]
        x_shape = [len(x_ax) for x_ax in x]

        model_0_values = {
            model_param_name: param_values[new_param_name]
            for model_param_name, new_param_name in self.model_param_maps[0].items()
        }
        model_1_values = {
            model_param_name: param_values[new_param_name]
            for model_param_name, new_param_name in self.model_param_maps[1].items()
        }

        y = np.zeros((self.get_num_y_channels(), x_shape[0], x_shape[1]))

        y_1 = np.atleast_2d(self.models[1].func(x_ax_1, model_1_values))

        for x_idx in range(len(x_ax_1)):
            model_0_values.update(
                {
                    param_name: y_1[param_idx, x_idx]
                    for param_idx, param_name in enumerate(self.result_params)
                }
            )
            y_idx = np.atleast_2d(self.models[0].func(x_ax_0, model_0_values))
            y[:, :, x_idx] = y_idx

        return y

    def estimate_parameters(self, x: TX2D, y: TY2D):
        x_ax_0 = x[0]
        x_ax_1 = x[1]

        # Step 1: find heuristics for model 0
        # Run the heuristics for each point on the second axis, then pick the best set
        heuristics = {
            param_name: np.zeros(len(x_ax_1) + 1)
            for param_name in self.models[0].parameters.keys()
        }

        for x_1_idx in range(len(x_ax_1)):
            self.models[0].clear_heuristics()
            self.models[0].estimate_parameters(x=x_ax_0, y=y[:, x_1_idx, :])
            for param_name, param_data in self.models[0].parameters.items():
                heuristics[param_name][x_1_idx] = param_data.get_initial_value()

        for param_name in heuristics.keys():
            heuristics[param_name][-1] = np.mean(heuristics[param_name][:-1])

        costs = np.zeros(len(x_ax_1) + 1)
        for idx in range(len(costs)):
            y_idx = self.models[0].func(
                x=x_ax_0,
                param_values={
                    param_name: heuristics[param_name][idx]
                    for param_name in heuristics.keys()
                },
            )
            costs[idx] = np.sqrt(np.sum((y - y_idx) ** 2))

        best_heuristic = np.argmin(costs)

        for model_param_name, new_param_name in self.model_param_maps[0].items():
            heuristic = heuristics[model_param_name][best_heuristic]
            self.parameters[new_param_name].heuristic = heuristic

        # Step 2: use the heuristics from model 0 as an input for model 1
        y_1 = np.zeros((len(self.result_params), len(x_ax_1)))
        for idx, result_param_name in enumerate(self.result_params):
            y_1[idx, :] = heuristics[result_param_name][:-1]

        self.models[1].estimate_parameters(x_ax_1, y_1)

        for model_param_name, new_param_name in self.model_param_maps[1].items():
            param_data = self.models[1].parameters[model_param_name]
            heuristic = param_data.get_initial_value()
            self.parameters[new_param_name].heuristic = heuristic

    def calculate_derived_params(
        self,
        x: TX2D,
        y: TY2D,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_values = {}
        derived_uncertainties = {}

        model_values = ({}, {})
        model_uncertainties = ({}, {})
        model_derived_values = ({}, {})
        model_derived_uncertainties = ({}, {})

        # Result parameters do not have meaningful fitted values / uncertainties
        for param_name in self.result_params:
            model_values[0][param_name] = float("nan")
            model_uncertainties[0][param_name] = float("nan")

        for model_idx, model in enumerate(self.models):
            param_map = self.model_param_maps[model_idx]
            model_values[model_idx].update(
                {
                    model_param_name: fitted_params[new_param_name]
                    for model_param_name, new_param_name in param_map.items()
                }
            )
            model_uncertainties[model_idx].update(
                {
                    model_param_name: fit_uncertainties[new_param_name]
                    for model_param_name, new_param_name in param_map.items()
                }
            )
            derived = model.calculate_derived_params(
                x=x[model_idx],
                y=np.full_like(x[model_idx], float("nan")),
                fitted_params=model_values[model_idx],
                fit_uncertainties=model_uncertainties[model_idx],
            )

            model_derived_values[model_idx].update(derived[0])
            model_derived_uncertainties[model_idx].update(derived[1])

        duplicates = set(model_derived_values[0].keys()).intersection(
            set(model_derived_values[1].keys())
        )
        if duplicates:
            raise ValueError(
                "Duplicate derived result names found between the two models: "
                f"{duplicates}."
            )

        derived_values = {}
        derived_values.update(model_derived_values[0])
        derived_values.update(model_derived_values[1])

        derived_uncertainties = {}
        derived_uncertainties.update(model_derived_uncertainties[0])
        derived_uncertainties.update(model_derived_uncertainties[1])

        return derived_values, derived_uncertainties


class Fitter2D:
    """Hierarchical fits to :class Model2D:s.

    Fitting proceeds as follows:
      - We fit the complete y-axis dataset to a :class RepeatedModel: based on the first
        of the :class Model2D:'s models. This performs a joint fit to all points on the
        second x-axis dimension.
      - The "result parameters" from the first model are used to create a second y-axis
        dataset, which the :class Model2D:'s second model is fit to.

    See :class Model2D: for more details. For example usage, see
    `test\test_multi_fitter`.
    """

    x: TX2D
    y: TY2D
    sigma: TY2D
    values: Dict[str, float]
    uncertainties: Dict[str, float]
    derived_values: Dict[str, float]
    derived_uncertainties: Dict[str, float]
    initial_values: Dict[str, float]
    model: Model2D
    free_parameters: List[str]

    def __init__(
        self,
        x: TX2D,
        y: TY2D,
        model: Model2D,
        fitter_cls: Optional[Fitter] = None,
        fitter_args: Optional[Dict] = None,
        **kwargs,
    ):
        """
        :param x: x-axis data in the for of a tuple `(x_axis_0, x_axis_1)`
        :param y: y-axis input data in the form of an array:
          shaped as `[y_channels, x_axis_0, x_axis_1]`
        :param fitter_cls: optional fitter class to use for the first fit (the second
          fit is always a normal fit). Defaults to :class NormalFitter:.
        :param fitter_args: optional dictionary of keyword arguments to pass into the
          fitter class
        """
        x = list(x)
        self.x = x = tuple([np.array(x_ax) for x_ax in x])
        self.y = y = np.array(y, ndmin=3)

        self.model = copy.deepcopy(model)
        self.fitter_cls = fitter_cls or NormalFitter
        self.fitter_args = dict(fitter_args or {})

        if len(self.x) != 2:
            raise ValueError("Fitter2D requires 2 x-axes")

        if len(self.x[0]) != self.y.shape[1] or len(self.x[1]) != self.y.shape[2]:
            raise ValueError("x-axis and y axis shapes must match")

        x_ax_0, x_ax_1 = self.x

        # fit along first x axis
        model_0 = self.model.models[0]
        common_params = [
            param_name
            for param_name in model_0.parameters
            if param_name not in self.model.result_params
        ]
        model_0 = RepeatedModel(
            model=self.model.models[0],
            common_params=common_params,
            num_repetitions=len(x_ax_1),
            aggregate_results=True,
        )

        y_0 = np.moveaxis(self.y, -1, 0)
        y_0 = np.reshape(y_0, (np.prod(y_0.shape[0:2]), y_0.shape[2]))

        if self.fitter_args.get("sigma") is not None:
            sigma_0 = self.fitter_args["sigma"]
            sigma_0 = np.array(sigma_0, ndmin=3)
            sigma_0 = np.moveaxis(sigma_0, -1, 0)
            sigma_0 = np.reshape(
                sigma_0, (np.prod(sigma_0.shape[0:2]), sigma_0.shape[2])
            )
            self.fitter_args["sigma"] = sigma_0

        fit_0 = self.fitter_cls(x=x_ax_0, y=y_0, model=model_0, **self.fitter_args)

        # aggregate results
        y_1 = np.zeros((len(self.model.result_params), len(x_ax_1)), dtype=np.float64)
        sigma_1 = np.zeros_like(y_1) if fit_0.sigma is not None else None

        for param_idx, param_name in enumerate(self.model.result_params):
            y_param = np.array(
                [fit_0.values[f"{param_name}_{idx}"] for idx in range(len(x_ax_1))]
            )
            sigma_param = np.array(
                [
                    fit_0.uncertainties[f"{param_name}_{idx}"]
                    for idx in range(len(x_ax_1))
                ]
            )

            y_1[param_idx, :] = y_param

            if sigma_1 is not None:
                sigma_1[param_idx, :] = sigma_param

        fit_1 = NormalFitter(
            x=x_ax_1,
            y=y_1,
            model=self.model.models[1],
            sigma=sigma_1,
        )

        # aggregate results
        self.fits = (fit_0, fit_1)
        self.values: Dict[str, float] = {}
        self.uncertainties: Dict[str, float] = {}
        self.derived_values: Dict[str, float] = {}
        self.derived_uncertainties: Dict[str, float] = {}
        self.initial_values: Dict[str, float] = {}
        self.free_parameters: List[str] = []

        for result_dict_name in ["values", "uncertainties", "initial_values"]:
            result_dict = getattr(self, result_dict_name)
            for model_idx, model in enumerate(self.model.models):
                fit_results = getattr(self.fits[model_idx], result_dict_name)
                param_map = self.model.model_param_maps[model_idx]
                result_dict.update(
                    {
                        new_param_name: fit_results[model_param_name]
                        for model_param_name, new_param_name in param_map.items()
                    }
                )

        for model_idx, model in enumerate(self.model.models):
            param_map = self.model.model_param_maps[model_idx]
            model_name = self.model.model_names[model_idx]
            suffix = f"_{model_name}" if model_name else ""
            fit = self.fits[model_idx]

            self.derived_values.update(
                {
                    f"{model_param_name}{suffix}": value
                    for model_param_name, value in fit.derived_values.items()
                }
            )
            self.derived_uncertainties.update(
                {
                    f"{model_param_name}{suffix}": value
                    for model_param_name, value in fit.derived_uncertainties.items()
                }
            )

            self.free_parameters += [
                param_map[model_param_name]
                for model_param_name, new_param_name in param_map.items()
                if model_param_name in fit.free_parameters
            ]

        self.sigma = fit_0.calc_sigma()
        if self.sigma is not None:
            y_shape = self.y.shape
            self.sigma = np.reshape(self.sigma, (y_shape[2], y_shape[0], y_shape[1]))
            self.sigma = np.moveaxis(self.sigma, 0, -1)

    def evaluate(
        self,
        plot_mode=False,
        x_fit: Optional[TX2D] = None,
    ) -> Union[
        Tuple[TX2D, TY2D],
        Tuple[
            Array[("num_samples_ax_0",), np.float64],
            Array[("num_samples_ax_1",), np.float64],
            Array[("num_samples_ax_1", "num_samples_ax_0"), np.float64],
        ],
    ]:
        """Evaluates the model function using the fitted parameter set.

        :param unpack: if True, we format the output data to be convenient for plotting
            with matplotlib.pcolormesh by splatting the x-axis list and squeezing +
            transposing the y-axis data.
        :param x_fit: optional x-axis points to evaluate the model at. If
            `None`, we use the values stored as attribute `x` of the fitter.

        :returns: tuple of x-axis values used and corresponding y-axis values
            of the fitted model
        """
        x_fit = x_fit if x_fit is not None else self.x
        y_fit = self.model(x_fit, **self.values)

        if plot_mode:
            return *x_fit, y_fit.squeeze().T

        return x_fit, y_fit

    def residuals(self) -> TY2D:
        """Returns an array of fit residuals."""
        return self.y - self.evaluate()[1]
